# SIMple Programming Language

from operator import add, mul, sub, eq, gt
global_frame = {'+':add, '*':mul, '-':sub,
                '==':lambda x,y:int(eq(x,y)),
                '>':lambda x,y:int(gt(x,y)),
                'block': lambda *args: args[-1],
                '__parent__':None}
builtins = list(global_frame.values())

def eval(cmd, frame = global_frame):
    if (not isinstance(cmd, list)) or len(cmd) == 1:
        return eval_single(cmd, frame)
    elif cmd[0] == '=':
        bind(cmd[1], eval(cmd[2], frame), frame)
    elif cmd[0] == 'def':
        bind(cmd[1],
             Function(cmd[1], cmd[2], cmd[3], frame),
             frame)
    elif cmd[0] == 'nonlocal':
        frame['__nonlocal__'].extend(cmd[1:])
    elif cmd[0] == 'if':
        return eval(cmd[2], frame) if eval(cmd[1], frame) \
            else eval(cmd[3], frame)
    else:
        op, *arg_list = [eval(arg, frame) for arg in cmd]
        if op in builtins:
            return op(*arg_list)
        else:
            return eval(op.code, op.make_frame(arg_list))

def eval_single(cmd, frame):
    if isinstance(cmd, list):
        return eval_single(cmd[0], frame)
    if cmd.isnumeric():
        return int(cmd)
    elif cmd in frame:
        return frame[cmd]
    else:
        return eval_single(cmd, frame['__parent__'])

class Function:
    def __init__(self, name, params, code, frame):
        self.name,self.params,self.code = name,params,code
        self.parent_frame = frame
    def make_frame(self, args_list):
        new_frame = dict(zip(self.params, args_list))
        new_frame['__parent__'] = self.parent_frame
        new_frame['__nonlocal__'] = []
        return new_frame
        
def bind(name, value, frame, redirected = False):
    if not redirected and ('__nonlocal__' not in frame \
       or name not in frame['__nonlocal__']) \
       or redirected and name in frame:
        frame[name] = value
    else:
        bind(name, value, frame['__parent__'], True)

def simpl_interpreter():
    from simpl_parser import simpl_parse, simpl_readline
    while True:
        try:
            cmd = simpl_readline()
            if cmd != '':
                result = eval(simpl_parse(cmd))
                if result != None:
                    print(result)
        except Exception as e:
            print('Nonsense in SimPL.')
